/*
 * $Id: main.cpp,v 1.20 2006/07/19 12:04:25 tsuruoka Exp $
 */

#include <stdio.h>
#include <fstream>
#include <map>
#include <list>
//#include "trie.h"
#include <iostream>
#include "maxent.h"
#include <sstream>

extern ME_Model_Data me_model_data[];

using namespace std;

const int NUM_TRAIN_SENTENCES = 99999999;
//const int NUM_TRAIN_SENTENCES = 2000;
const double BIAS_FOR_RECALL = 0.6;
//const double BIAS_FOR_RECALL = 0.5;
//const double BIAS_FOR_RECALL = -0.2;

const bool USE_TRIGRAM = false;

int max_term_length = 0;

struct Token
{
  string str;
  string pos;
  string chunk;
  string label;
  Token(string s, string p, string l) : str(s), pos(p), label(l) {}
};

struct WordInfo
{
  string str;
  int inside_ne;
  int edge_ne;
  int total;
  WordInfo() : str(""), inside_ne(0), total(0), edge_ne(0) {};
  WordInfo(const string & s, int i, int e, int t) : str(s), inside_ne(i), total(t), edge_ne(e) {};
  double out_prob() const {
    return ( (total - inside_ne) + 1.0 ) / ( total + 2.0 );
  }
  double in_prob() const {
    return ( inside_ne + 1.0 ) / ( total + 2.0 );
  }
  double edge_prob() const {
    return ( edge_ne + 1.0 ) / ( total + 2.0 );
  }
  bool operator<(const WordInfo & x) {
    //    return this->out_prob() > x.out_prob();
    return this->edge_prob() > x.edge_prob();
  }
};

struct FeatureBag
{
  map<int, int> int2id;
  vector<int> id2int;
  int Id(int i) {
    map<int, int>::const_iterator j = int2id.find(i);
    if (j == int2id.end()) {
      int id = id2int.size();
      id2int.push_back(i);
      int2id[i] = id;
      return id;
    }
    return j->second;
  }
  int Num(int id) const {
    assert(id >= 0 && id < id2int.size());
    return id2int[id];
  }
};

map<string, WordInfo> word_info;
map<string, WordInfo> pos_info;
map<string, WordInfo> trigram;

typedef vector<Token> Sentence;

int ambiguity_matrix[10][10];

string majority(const vector<string> & li)
{
  map<string, int> m;
  for (vector<string>::const_iterator i = li.begin(); i != li.end(); i++) {
    m[*i]++;
  }
  int max = 0;
  string maj = "";
  for (map<string, int>::const_iterator i = m.begin(); i != m.end(); i++) {
    if (i->second > max) {
      max = i->second;
      maj = i->first;
    }
    // count ambiguity
    //    for (map<int, int>::const_iterator j = m.begin(); j != m.end(); j++) {
    //      ambiguity_matrix[class2int(i->first)][class2int(j->first)]++;
    //    }
  }
  return maj;  
}


static bool isterm(const Sentence & s, const int begin, const int end)
{
  if (s[begin].label[0] != 'B') return false;
  for (int x = begin + 1; x < end; x++) {
    if (s[x].label[0] != 'I') return false;
  }
  if (end < s.size() && s[end].label[0] == 'I') return false;

  return true;
}

int trigrams(const string & s, list<string> & ls)
{
  if (s.size() <= 2) return 0;
  for (int i = 0; i < s.size() - 2; i++) {
    string t = s.substr(i, 3);
    ls.push_back(t);
  }
}

int makedic(const vector<Sentence> & vs,
            map<string, string> & dic)
{
  cerr << "making dictionary...";
  
  multimap<string, string> tmpdic;
  
  int nterms = 0;
  for (vector<Sentence>::const_iterator i = vs.begin(); i != vs.end(); i++) {
    const Sentence & s = *i;
    string whole;

    for (int j = 0; j < s.size(); j++) {
      word_info[s[j].str].str = s[j].str;
      word_info[s[j].str].total++;
      pos_info[s[j].pos].str = s[j].pos;
      pos_info[s[j].pos].total++;
      if (j == 0) whole = s[j].str;
      else        whole += " " + s[j].str;
    }
    if (USE_TRIGRAM) {
      list<string> tl;
      trigrams(whole, tl);
      for (list<string>::const_iterator m = tl.begin(); m != tl.end(); m++) {
        trigram[*m].str = *m;
        trigram[*m].total++;
      }
    }

    for (int j = 0; j < s.size(); j++) {
      if (s[j].label[0] == 'B') {
        string label = s[j].label.substr(2);
        string term = s[j].str;
        word_info[s[j].str].str = s[j].str;
        word_info[s[j].str].inside_ne++;
        word_info[s[j].str].edge_ne++;
        pos_info[s[j].pos].str = s[j].pos;
        pos_info[s[j].pos].inside_ne++;
        int k = j + 1;
        while (k < s.size()) {
          //          if (s[k].label[0] == 'O') break;
          if (s[k].label[0] != 'I') break;
          term += " " + s[k].str;
          word_info[s[k].str].str = s[k].str;
          word_info[s[k].str].inside_ne++;
          pos_info[s[k].pos].str = s[k].pos;
          pos_info[s[k].pos].inside_ne++;
          k++;
        }
        word_info[s[k-1].str].edge_ne++;
        if (k - j > max_term_length) max_term_length = k - j;
        //        cout << term << "\t" << label << endl;
        tmpdic.insert(pair<string, string>(term, label));

        if (USE_TRIGRAM) {
          list<string> tl;
          trigrams(term, tl);
          for (list<string>::const_iterator m = tl.begin(); m != tl.end(); m++) {
            trigram[*m].str = *m;
            trigram[*m].inside_ne++;
          }
        }

        nterms++;
        j = k - 1;
      }
    }
  }

  //  AString prestr(-1);
  string prestr = "#!!!#";
  int max_n = 0;
  vector<string> classes;
  int num_terms = 0;
  int num_ambiguous = 0;
  for (multimap<string, string>::const_iterator i = tmpdic.begin(); ; i++) {
    if (i == tmpdic.end() || i->first != prestr) {
      if (prestr != "#!!!#") {
        string maj = majority(classes);
        dic.insert(pair<string, string>(prestr, maj));
        //        cout << "#" << prestr << "\t" << maj << endl;

        // ambiguity count
        //        num_terms++;
        //        //        classes.sort();
        //        classes.unique();
        //        if (classes.size() > 1) num_ambiguous++;
      }
      classes.clear();
      if (i == tmpdic.end()) break;
    }
    //    cout << i->first << "\t" << i->second << endl;
    prestr = i->first;
    classes.push_back(i->second);
  }
  //  cout << "total number of terms = " << num_terms << endl;
  //  cout << "number of ambigous terms = " << num_ambiguous << endl;
  
  cerr << "done" << endl;
  cerr << "number of terms = " << nterms << endl;
  cerr << "max_term_length = " << max_term_length << endl;
}

static string normalize(const string & s)
{
  string tmp;
  for (int i = 0; i < s.size(); i++) {
    char c = tolower(s[i]);
    if (isdigit(c)) c = '#';
    if (c == '-' || c == ' ') continue;
    tmp += c;
  }
  /*
  if (tmp == "is") tmp = "be";
  if (tmp == "was") tmp = "be";
  if (tmp == "are") tmp = "be";
  if (tmp == "were") tmp = "be";
  if (tmp == "an") tmp = "a";
  if (tmp == "the") tmp = "a";
  */
  if (tmp[tmp.size()-1] == 's') return tmp.substr(0, tmp.size()-1);
  return tmp;
}

static string wordshape(const string & s, bool fine)
{
  string tmp;
  char pre_c = 0;
  for (int i = 0; i < s.size(); i++) {
    char c = s[i];
    if (isdigit(c)) c = '#';
    else if (isupper(c)) c = 'A';
    else if (islower(c)) c = 'a';
    else if (c == ' ' || c == '-') c = '-';
    else continue;
    if (fine || c != pre_c) 
      tmp += c;
    pre_c = c;
  }
  return tmp;
}

ME_Sample mesample(const FeatureBag & fb,
                   const string & label, const vector<Token> &vt, int begin, int end)
{
  ME_Sample mes;

  mes.label = label;

  char buf[1000];

  string s;

  // contextual feature
  string s_1, s_2, s1, s2;
  //  if (begin >= 1) s_1 = vt[begin-1].str;
  if (begin >= 1) s_1 = normalize(vt[begin-1].str);
  else            s_1 = "BOS";
  sprintf(buf, "C-1_%s", s_1.c_str());
  mes.features.push_back(buf);

  //  if (end < vt.size()) s1 = vt[end].str;
  if (end < vt.size()) s1 = normalize(vt[end].str);
  else                 s1 = "EOS";
  sprintf(buf, "C+1_%s", s1.c_str());
  mes.features.push_back(buf);

  //  if (begin >= 2) s_2 = vt[begin-2].str;
  if (begin >= 2) s_2 = normalize(vt[begin-2].str);
  else            s_2 = "BOS2";
  sprintf(buf, "C-2_%s", s_2.c_str());
  //  mes.features.push_back(buf);

  //  if (end < vt.size()-1) s2 = vt[end+1].str;
  if (end < vt.size()-1) s2 = normalize(vt[end+1].str);
  else                   s2 = "EOS2";
  sprintf(buf, "C+2_%s", s2.c_str());
  //  mes.features.push_back(buf);

  mes.features.push_back("C-2-1_" + s_2 + "_" + s_1);
  mes.features.push_back("C-1+1_" + s_1 + "_" + s1);
  mes.features.push_back("C+1+2_" + s1  + "_" + s2);

  /*
  // bag of words
  //  for (int i = begin - 5; i <= end -1 + 5; i++) {
  for (int i = begin - 4; i <= end -1 + 4; i++) {
    if (i >= 0 && i < vt.size() && (i < begin || i >= end)) {
      sprintf(buf, "CSEN_%s", normalize(vt[i].str).c_str());
      mes.features.push_back(buf);
    }
  }
  */
  
  // term feature
  char firstletter = vt[begin].str[0];
  char lastletter = vt[end-1].str[vt[end-1].str.size()-1];
  //  if (begin != 0 && isupper(firstletter))
  //  if (isupper(firstletter) && isupper(lastletter)) 
  //    mes.features.push_back("IS_UPPER");

  //  if (end - begin == 1) {
  //    mes.features.push_back("EXACT_" + vt[begin].str);
  //  }
  
  string tb = normalize(vt[begin].str);
  sprintf(buf, "TB_%s", tb.c_str());
  mes.features.push_back(buf);
  
  for (int i = begin + 1; i < end-1; i++) {
  //for (int i = begin; i < end; i++) {
    s = normalize(vt[i].str);
    sprintf(buf, "TM_%s", s.c_str());
    mes.features.push_back(buf);
  }

  string te = normalize(vt[end-1].str);
  sprintf(buf, "TE_%s", te.c_str());
  mes.features.push_back(buf);

  /*
  if (end - begin > 1) {
    string tb2 = normalize(vt[begin+1].str);
    //    sprintf(buf, "TB2_%s", tb2.c_str());
    sprintf(buf, "TB12_%s_%s", tb.c_str(), tb2.c_str());
    mes.features.push_back(buf);
    string te2 = normalize(vt[end-2].str);
    //    sprintf(buf, "TE2_%s", te2.c_str());
    sprintf(buf, "TE21_%s_%s", te.c_str(), te2.c_str());
    mes.features.push_back(buf);
  }
  */
  
  // combination
  mes.features.push_back("C-1_TB_" + s_1 + "_" + tb);
  mes.features.push_back("C-1_TE_" + s_1 + "_" + te);
  mes.features.push_back("TB_TE_" + tb + "-" + te);
  mes.features.push_back("TB_C+1_" + tb + "_" + s1);
  mes.features.push_back("TE_C+1_" + te + "-" + s1);

  //  mes.features.push_back("C-2-1_TE_" + s_2 + "_" + s_1 + "_" + te);
  //  mes.features.push_back("TE_C+1+2_" + te + "_" + s1 + "_" + s2);
  
  // length
  if (end - begin > 5) 
    mes.features.push_back("LEN>=5");
  else {
    sprintf(buf, "LEN=%d", end - begin);
    mes.features.push_back(buf);
  }
  
  // character trigram
  if (USE_TRIGRAM) {
    string exact_whole = "";
    for (int i = begin; i < end; i++) {
      if (i == begin) exact_whole = vt[i].str;
      else            exact_whole += " " + vt[i].str;
    }
    list<string> tl;
    trigrams(exact_whole, tl);
    for (list<string>::const_iterator m = tl.begin(); m != tl.end(); m++) {
      //      if (trigram[*m].in_prob() > 0.9)
      if (trigram[*m].out_prob() > 0.9)
        mes.features.push_back("TRIGRAM_" + *m);
    }
  }

  s = "";
  string whole = "";
  bool contain_comma = false;
  for (int i = begin; i < end; i++) {
    s += normalize(vt[i].str);
    whole += vt[i].str;
  }
  sprintf(buf, "WHOLE_%s", s.c_str());
  //if (label > 0) mes.features.push_back(buf);
  mes.features.push_back(buf);
  mes.features.push_back("WS1_" + wordshape(whole, true));
  mes.features.push_back("WS2_" + wordshape(whole, false));

  //  mes.features.push_back("WHOLE_C+1_" + whole + "-" + s1);
  
  // dictionary matching
  /*
  string dicstr = "";
  for (int i = begin; i < end; i++) {
    if (i == begin) dicstr = vt[i].str;
    else            dicstr += vt[i].str;
  }
  const map<string, string>::const_iterator idx = dic.find(dicstr);
  if (idx != dic.end()) {
    mes.positive_features.push_back(fb.Id(string("DICMATCH_" + idx->second)));
  }
  */

  // preffix and suffix
  //  if (s.size() >= 2) mes.features.push_back(fb.Id(string("WHOLE(pre2)_"+s.substr(0, 2))));
  if (s.size() >= 2) mes.features.push_back("WHOLE(suf2)_"+s.substr(s.size() - 2));
  if (s.size() >= 3) mes.features.push_back("WHOLE(pre3)_"+s.substr(0, 3));
  if (s.size() >= 3) mes.features.push_back("WHOLE(suf3)_"+s.substr(s.size() - 3));
  //  if (s.size() >= 4) mes.features.push_back("WHOLE(pre4)_"+s.substr(0, 4));
  if (s.size() >= 4) mes.features.push_back("WHOLE(suf4)_"+s.substr(s.size() - 4));
  //if (s.size() >= 7) mes.features.push_back("WHOLE(pre3+suf3)_"+s.substr(0, 3) + "_" + s.substr(s.size() - 3));

  // coordination
  /*
  bool comma = false;
  bool coordination = false;
  for (int i = begin + 1; i < end - 1; i++) {
    if (vt[i].str == ",")  comma = true;
    if (comma && vt[i].str == "and") coordination = true;
  }
  if (coordination)
    mes.features.push_back("COORDINATION");
  */
  
  //  if (contain_comma)
  //    mes.features.push_back("CONTAIN_COMMA");

  //  cout << fb.Id(string(buf)) << " " << string(buf) << endl;

  /*
  // chunk feature
  string c_2 = "BOS2", c_1 = "BOS1";
  string cb, ce;
  string c1 = "EOS1", c2 = "EOS2";
  if (begin >= 2) c_2 = vt[begin-2].chunk;
  if (begin >= 1) c_1 = vt[begin-1].chunk;
  cb = vt[begin].chunk;
  ce = vt[end-1].chunk;
  if (end < vt.size())   c1 = vt[end].chunk;
  if (end < vt.size()-1) c2 = vt[end+1].chunk;

  mes.features.push_back("Chk-1_" + c_1);
  mes.features.push_back("Chk-B_" + cb);
  mes.features.push_back("Chk-E_" + ce);
  mes.features.push_back("Chk+1_" + c1);
  */

  /*
  // POS feature
  string p_2 = "BOS2", p_1 = "BOS1";
  string pb, pe;
  string p1 = "EOS1", p2 = "EOS2";
  if (begin >= 2) p_2 = vt[begin-2].pos;
  if (begin >= 1) p_1 = vt[begin-1].pos;
  pb = vt[begin].pos;
  pe = vt[end-1].pos;
  if (end < vt.size())   p1 = vt[end].pos;
  if (end < vt.size()-1) p2 = vt[end+1].pos;

  mes.features.push_back("PoS-1_" + p_1);
  mes.features.push_back("PoS-B_" + pb);
  mes.features.push_back("PoS-E_" + pe);
  mes.features.push_back("PoS+1_" + p1);

  //  string posseq;
  //  for (int i = begin; i < end; i++) {
  //    posseq += vt[i].pos + "_";
  //  }
  //  mes.features.push_back("PosSeq_" + posseq);
  
  //  for (int i = begin + 1; i < end-1; i++) {
  //    sprintf(buf, "PoS-M_%s", vt[i].pos.c_str());
  //    mes.features.push_back(buf);
  //  }
  
  //  mes.features.push_back("PoS-2-1_" + p_2 + "_" + p_1);
  //  mes.features.push_back("PoS-1+1_" + p_1 + "_" + p1);
  //  mes.features.push_back("PoS+1+2_" + p1 + "_" + p2);
  //mes.features.push_back("PoS-1B_" + p_1 + "_" + pb);
  //mes.features.push_back("PoSE+1_" + pe + "_" + p1);
  //  mes.features.push_back("PoSBE_" + pb + "_" + pe);
  
  mes.features.push_back("PoS-1BE+1_" + p_1 + "_" + pb + "_" + pe + "_" + p1);
  */
  
  /*
  cout << "label = " << label << "\t";
  for (int i = begin; i < end; i++) {
    cout << vt[i].str << " ";
  }
  cout << endl;
  for (list<string>::const_iterator i = mes.features.begin(); i != mes.features.end(); i++){
    cout << *i << " ";
  }
  cout << endl << endl;
  */
  
  return mes;
}

static bool
is_inside_np(const Sentence & s, const int begin, const int end)
{
  if (s[begin].chunk != "B-NP" && s[begin].chunk != "I-NP") return false;

  bool inside_paren = false;
  for (int i = begin + 1; i < end; i++) {
    const string & c = s[i].chunk;
    const string & str = s[i].str;
    const string & p = s[i].pos;
    if (p == ")") { inside_paren = false; continue; }
    if (p == "(") { inside_paren = true; continue; }
    if (inside_paren) continue;
    //    if (c != "I-NP" && c != "B-NP") return false;
    if (c != "I-NP" && c != "B-NP" && p != "CC" && p != "," &&
        str != "of" && str != "to") return false;
  }
  return true;
}

static bool is_candidate(const Sentence & s, const int begin, const int end)
{
  /*
  bool inside_np = is_inside_np(s, begin, end);
  if (!inside_np && isterm(s, begin, end)) {
    for (int j = begin; j < end; j++) cout << s[j].str << "\t" << s[j].chunk << "\t" << s[j].label << endl;
    cout << endl;
  }
  return inside_np;
  */

  
  //  if ((end - begin) > max_term_length ) return false;
  /*
  int penalty = 0;
  for (int x = begin; x < end; x++) {
    double out_prob = word_info[s[x].str].out_prob();
    if (out_prob >= 0.98) return false;
    out_prob = pos_info[s[x].pos].out_prob();
    if (out_prob >= 0.98) return false;
  }
  */
  if (word_info[s[begin].str].edge_prob() < 0.01) return false;
  if (word_info[s[end-1].str].edge_prob() < 0.01) return false;
  //  if (end - begin > 10) return false;

  int penalty = 0;
  int kakko = 0;
  for (int x = begin; x < end; x++) {
    //    if (s[x].str == string("(")) kakko++;
    //    if (s[x].str == string(")")) kakko--;
    double out_prob = word_info[s[x].str].out_prob();
    //    if (out_prob >= 0.99) penalty++;
    //    if (out_prob >= 0.90) penalty++;
    //    if (out_prob >= 0.98) penalty++;
    //    if (out_prob >= 0.94) penalty++;
    if (out_prob >= 0.99) penalty++;
    if (out_prob >= 0.97) penalty++;
    //    if (penalty >= 2) return false;
    if (penalty >= 3) return false;
  }

  if (kakko % 2 != 0) return false;
  
    /*
  double prod = 1.0;
  for (int x = begin; x < end; x++) {
    map<string, WordInfo>::const_iterator i = word_info.find(s[x].str);
    if (i != word_info.end())
      prod *= i->second.in_prob();
    //    cout << s[x].str << ":" << prod << " ";
    if (prod < 0.0001) {
      if (isterm(s, begin, end)) cerr << "*";
      return false;
    }
  }
    */
  return true;
}

void load_word_info(const string & filename)
{
  ifstream ifile(filename.c_str());
  if (!ifile) { cerr << "error: cannot open " << filename << endl; exit(1); }

  word_info.clear();
  string line;
  while (getline(ifile, line)) {
    istringstream is(line.c_str());
    string s;
    int i, e, t;
    is >> s >> i >> e >> t;
    WordInfo wi(s, i, e, t);
    word_info.insert(make_pair(s, wi));
  }
}

void save_word_info(const string & filename)
{
  ofstream ofile(filename.c_str());
  if (!ofile) { cerr << "error: cannot open " << filename << endl; exit(1); }

  for (map<string, WordInfo>::const_iterator i = word_info.begin(); i != word_info.end(); i++) {
    const WordInfo & wi = i->second;
    ofile << wi.str << " " << wi.inside_ne << " " << wi.edge_ne << " " << wi.total << endl;
  }

}

int train(const vector<Sentence> & vs, const map<string, string> & dic,
          ME_Model & me, FeatureBag & fb, double width_factor)
{
  cerr << "extracting features...";
  int n = 0;
  int nterms = 0;
  for (vector<Sentence>::const_iterator i = vs.begin(); i != vs.end(); i++) {
    const Sentence & s = *i;
    for (int j = 0; j < s.size(); j++) {
      for (int k = j + 1; k <= s.size(); k++) {
        if (is_candidate(s, j, k)) {
          string label = "O";
          if (isterm(s, j, k)) {
            label = s[j].label.substr(2);
            nterms++;
          }
          me.add_training_sample(mesample(fb, label, s, j, k));
        }  else {
          //          break;
        }
      }
    }
  }
  cerr << "done" << endl;
  cerr << "number of train terms = " << nterms << endl;

  me.set_heldout(1000, 0);
  //  me.learn(train_samples, 5000, 0, 0, 1.0 / 1);
  //  me.train(train_samples, 0, 500);
  //  me.train(train_samples, 0, 0, 1.0);
  me.train(0, 0, width_factor);
  //  me.train(0, 500);
  //  me.train(train_samples, 0, 0, 0.5);

  /*
  // print ME features
  list< pair< pair<string, string>, double> > fl;
  me.get_features(fl);
  for (list< pair< pair<string, string>, double> >::const_iterator i = fl.begin(); i != fl.end(); i++) {
    printf("%10.3f  %-10s %s\n", i->second, i->first.first.c_str(), i->first.second.c_str());
  }
  cout << endl;
  */
}

void print_features(const FeatureBag & fb, const ME_Sample & nbs)
{
  cout << endl;
  for (vector<string>::const_iterator i = nbs.features.begin(); i != nbs.features.end(); i++) {
    cout << *i << " ";
  }
}


struct Annotation {
  int label;
  int begin;
  int end;
  double prob;
  bool operator<(const Annotation & x) { return prob > x.prob; }
  Annotation(const int l, const int b, const int e, const double p) :
    label(l), begin(b), end(e), prob(p) {}
};

/*
bool is_term(const Sentence & s, int j, int k)
{
  if (s[j].label.substr(0, 2) != "B-") return false;
  string cat = s[j].label.substr(2);
  for (int i = j + 1; i <= k; i++)
    if (s[j].label.substr(0, 2) != "I-" || s[j].label.substr(2) != cat) return false;

  if (k == s.size() - 1) return true;
  if (s[k+1].label == "O" || s[k+1].label.substr(0, 2) == "B-") return true;
  return false;
}
*/

void find_NEs(const FeatureBag & fb, const ME_Model & me,
              Sentence & s)
{
  const int other_class = me.get_class_id("O");
  
  vector<double> label_p(s.size());
  for (int j = 0; j < s.size(); j++) {
    s[j].label = string("O");
    label_p[j] = 0;
  }
  list<Annotation> la;
  for (int j = 0; j < s.size(); j++) {
    //      for (int k = s.size(); k > j; k--) {
    for (int k = j + 1; k <= s.size(); k++) {
      if (!is_candidate(s, j, k)) {
        //        if (isterm(s_org, j, k)) num_candidate_false_negatives++;
        continue;
      }
      ME_Sample nbs = mesample(fb, "?", s, j, k);
      vector<double> membp(me.num_classes());
      //        int label = nb.classify(nbs, NULL, &membp);
      //        me.classify(nbs, &membp);
      membp = me.classify(nbs);
      int label = 0;
      membp[other_class] -= BIAS_FOR_RECALL;
      for (int l = 0; l < me.num_classes(); l++) {
        if (membp[l] > membp[label]) label = l;
      }
      double prob = membp[label];
      /*
        print_features(fb, nbs);
        cout << endl << "------- ";
        for (int l = 0; l < me.num_classes(); l++) cout << membp[l] << " ";
        cout << endl;
      */
      if (label != other_class) {
        la.push_back(Annotation(label, j, k, prob));
      }
    }
  }
  la.sort();
  //    for (int j = 0; j < s.size(); j++) cout << j << ":" << s[j].str << " ";
  //    cout << endl;
  for (list<Annotation>::const_iterator j = la.begin(); j != la.end(); j++) {
    //      cout << j->label << " begin = " << j->begin << " end = " << j->end << " prob = " << j->prob << endl;
    bool override = true;
    for (int l = j->begin; l < j->end; l++) {
      if (label_p[l] >= j->prob) { override = false; break; }
      if (s[l].label != string("O")) {
        // erase the old label
        int lbegin = l;
        while (s[lbegin].label[0] != 'B') lbegin--;
        int lend = l;
        while (s[lend].label[0] != 'O' && lend < s.size()) lend++;
        for (int t = lbegin; t < lend; t++) {
          s[t].label = string("O");
          label_p[t] = 0;
        }
      }
    }
    if (!override) continue;
    for (int l = j->begin; l < j->end; l++) {
      label_p[l] = j->prob;
      if (l == j->begin)  s[l].label = "B-" + me.get_class_label(j->label);
      else                s[l].label = "I-" + me.get_class_label(j->label);
    }
  }


}


void mlmatch(FeatureBag & fb,
             vector<Sentence> & vs,
             ME_Model & me)
{
  int num_candidate_false_negatives = 0;
  for (vector<Sentence>::iterator i = vs.begin(); i != vs.end(); i++) {
    find_NEs(fb, me, *i);
    // print
    for (int j = 0; j < i->size(); j++) {
      cout << (*i)[j].str << "\t" << (*i)[j].label << endl;
    }
    cout << endl;
    
  }
  //  cerr << "num_candidate_false_negatives = " << num_candidate_false_negatives << endl;

}
/*
void dicmatch(vector<Sentence> & vs,
              const map<string, string> & dic)
{
  Trie trie;
  for (map<string, string>::const_iterator i = dic.begin(); i != dic.end(); i++) {
    trie.Insert(i->first);
  }
  
  // longest matching
  for (vector<Sentence>::iterator i = vs.begin();
       i != vs.end(); i++) {
    //    cerr << ".";
    Sentence & s = *i;
    for (int j = 0; j < s.size(); j++) {
      s[j].label = string("O");
    }
    for (int j = 0; j < s.size(); j++) {
      for (int k = s.size(); k > j; k--) {
        string term = s[j].str;
        for (int l = j + 1; l < k; l++) {
          term += " " + s[l].str;
        }
        if (!trie.Lookup(term)) continue;
        const map<string, string>::const_iterator idx = dic.find(term);
        if (idx != dic.end()) {
          for (int l = j; l < k; l++) {
            if (l == j)  s[l].label = "B-" + idx->second;
            else         s[l].label = "I-" + idx->second;
          }
          j = k - 1;
        }
        
      }
    }
  }
}
*/

/*
void write_iob2(const string & filename, vector<Sentence> & vs)
{
  ofstream ofile(filename.c_str());

  for (vector<Sentence>::const_iterator i = vs.begin();
       i != vs.end(); i++) {
    const Sentence & s = *i;
    for (int j = 0; j < s.size(); j++) {
      ofile << s[j].str << "\t" << s[j].label << endl;
    }
    ofile << endl;
  }
  ofile.close();
}
*/
 
void read_pos_iob2(const string & filename, vector<Sentence> & vs, int num_sentences = -1)
{
  ifstream ifile(filename.c_str());

  string line;
  Sentence sentence;
  int n = 0;
  cerr << "reading " << filename;
  while (getline(ifile,line)) {
    if (line.size() == 0) {
      vs.push_back(sentence);
      if (num_sentences > 0 && vs.size() == num_sentences) break;
      sentence.clear();
      continue;
    }
    char buf1[256], buf2[256], buf3[256];
    sscanf(line.c_str(), "%s\t%s\t%s", buf1, buf2, buf3);
    //    cout << buf1 << " " << buf2 << " " << buf3 << endl;
    Token token(buf1, buf2, buf3);
    sentence.push_back(token);
    if (n++ % 10000 == 0) cerr << ".";
  }
  cerr << endl;

  ifile.close();
}

void read_merged(const string & filename, vector<Sentence> & vs, int num_sentences = -1)
{
  ifstream ifile(filename.c_str());

  string line;
  Sentence sentence;
  int n = 0;
  cerr << "reading " << filename;
  while (getline(ifile,line)) {
    if (line.size() == 0 || line[0] == '\t') {
      vs.push_back(sentence);
      if (num_sentences > 0 && vs.size() == num_sentences) break;
      sentence.clear();
      continue;
    }
    istringstream is(line.c_str());
    string str, base, pos, chunk, label;
    is >> str >> base >> pos >> chunk >> label;
    Token token(str, pos, label);
    //    cout << str << " " << pos << " " << label << endl;
    token.chunk = chunk;
    sentence.push_back(token);
    if (n++ % 10000 == 0) cerr << ".";
  }
  cerr << endl;

  ifile.close();
}


int main(int argc, char** argv)
{
  double width_factor = 1.0;
  if (argc == 2) {
    width_factor = atof(argv[1]);
  }
  
  //  const string train_file = "./Genia4ERtask1.pos.iob2";
  //  const string test_file  = "./Genia4EReval1.pos.iob2";
  const string train_file = "./train.merged";
  //  const string train_file = "./train.pos.iob2";
  const string test_file  = "./dev.merged";
  //    const string test_file  = "./train.iob2";

  vector<Sentence> trains, tests;
  map<string, string> dic;

  //  for (int i = 0; i < 10; i++) 
  //    for (int j = 0; j < 10; j++)
  //      ambiguity_matrix[i][j] = 0;
  
  //  read_pos_iob2(train_file, trains, NUM_TRAIN_SENTENCES);
  /*
  read_merged(train_file, trains, NUM_TRAIN_SENTENCES);
  makedic(trains, dic);
  save_word_info("word_info");
  */
  
  /*
  list<WordInfo> lwi;
  for (map<string, WordInfo>::const_iterator i = word_info.begin(); i != word_info.end(); i++) {
  //  for (map<string, WordInfo>::const_iterator i = trigram.begin(); i != trigram.end(); i++) {
    //    cout << i->first << "\t" << i->second.inside_ne << "/" << i->second.total << endl;
    lwi.push_back(i->second);
  }
  lwi.sort();
  for (list<WordInfo>::const_iterator i = lwi.begin(); i != lwi.end(); i++) {
    cout << i->str << "\t" << i->edge_ne << "/" << i->total << " = " << i->edge_prob() << endl;
  }
  exit(0);
  */
  
  FeatureBag fb;
  ME_Model me;
  char buf[100];
  sprintf(buf, "model_%4.2f", width_factor);
  //  train(trains, dic, me, fb, width_factor);
  //  me.save_to_file(buf);
  me.load_from_file("model_1.00");
  //  me.load_from_array(me_model_data);
  load_word_info("word_info");


  string line;
  while(getline(cin, line)) {
    istringstream is(line.c_str());
    Sentence s;
    string str;
    while(is >> str) {
      Token t(str, "", "");
      s.push_back(t);
    }
    find_NEs(fb, me, s);
    for (int j = 0; j < s.size(); j++) {
      cout << s[j].str << "\t" << s[j].label << endl;
    }
    cout << endl;
  }
  exit(0);

  
  read_merged(test_file, tests);
  //  read_pos_iob2(test_file, tests);
  //  read_iob2(train_file, tests, NUM_TRAIN_SENTENCES);
  //  dicmatch(tests, dic);
  mlmatch(fb, tests, me);
  //  char buf[100];
  //  sprintf(buf, "results_%4.2f.iob2", width_factor);
  //  write_iob2(buf, tests);
  
}

/*
 * $Log: main.cpp,v $
 * Revision 1.20  2006/07/19 12:04:25  tsuruoka
 * add wordshape
 *
 * Revision 1.19  2004/08/27 13:27:17  tsuruoka
 * get_class_name() -> get_class_label()
 *
 * Revision 1.18  2004/07/26 13:08:47  tsuruoka
 * remove maxent
 *
 * Revision 1.17  2004/07/21 10:59:44  tsuruoka
 * add inequality ME
 *
 * Revision 1.16  2004/07/17 10:11:11  tsuruoka
 * modify mlmatch algorithm
 *
 * Revision 1.15  2004/07/11 06:11:16  tsuruoka
 * remove int2class
 *
 * Revision 1.14  2004/07/10 14:41:39  tsuruoka
 * modify BIAS algorithm
 *
 * Revision 1.13  2004/07/10 05:58:28  tsuruoka
 * add edge_prob()
 *
 * Revision 1.12  2004/07/08 11:43:12  tsuruoka
 * add early stopping
 *
 * Revision 1.11  2004/07/07 11:38:33  tsuruoka
 * add dic_feature
 *
 * Revision 1.10  2004/07/04 09:47:48  tsuruoka
 * add POS
 *
 * Revision 1.9  2004/07/04 06:26:49  tsuruoka
 * add trigram,.. etc
 *
 * Revision 1.8  2004/07/03 15:42:22  tsuruoka
 * modify is_candidate
 *
 * Revision 1.7  2004/07/03 02:06:54  tsuruoka
 * add features
 *
 * Revision 1.6  2004/07/02 09:15:36  tsuruoka
 * add LMVM
 *
 * Revision 1.5  2004/07/02 06:59:05  tsuruoka
 * add combination features
 *
 * Revision 1.4  2004/06/30 14:17:18  tsuruoka
 * add get_features()
 *
 * Revision 1.3  2004/06/30 09:16:16  tsuruoka
 * add maxent
 *
 * Revision 1.2  2004/06/27 11:03:11  tsuruoka
 * add nblearner
 *
 */

