#include "nblearner.h"
#include <stdexcept>
#include <fstream>
#include <math.h>
#include <iostream>

using namespace std;

//vector<string> split(const string& s);

const double DEFAULT_DELTA = 0.0001;
//const double DELTA = 0.0001;
//const double DELTA = 0.001;
//const double DELTA = 0.01;

struct Node {
  vector<bool> used_features;
  double precision;
};


//------------------------------------------------------------------------
// optimize DELTA
//     ... optimize the value of DELTA using leave_one_out()
//------------------------------------------------------------------------
void NB_Learner::optimize_delta()
{
  double delta = 0.001;
  //  double delta = 0.1;
  double coeff = 0.1;

  double pre_precision = 0;
  while(1) {
    DELTA = delta;
    double precision = fast_leave_one_out();
    cerr << "DELTA = " << delta << " precision = " << precision << endl;
    //    if (precision <= pre_precision || delta < 0.0000001) {
    if (precision < pre_precision || delta < 0.000000001) {
      DELTA = delta / coeff;
      break;
    }
    delta *= coeff;
    pre_precision = precision;
  }


}


//------------------------------------------------------------------------
// Disable features with low frequency
//------------------------------------------------------------------------
void NB_Learner::feature_count(int count)
{
  for (int j = 0; j < _num_features; j++) {
    int sum = 0;
    for (int i = 0; i < _num_classes; i++) {
      sum += _cv[i][j];
    }
    if (sum >= count) _is_used[j] = true;
    else              _is_used[j] = false;
  }
  return;
}


//------------------------------------------------------------------------
// Feature selection
//------------------------------------------------------------------------
void NB_Learner::feature_selection()
{
  vector<Node> nv;

  Node n;
  for (int i = 0; i < _num_features; i++) n.used_features.push_back(true);
  _is_used = n.used_features;
  n.precision = fast_leave_one_out();
  //  cout << n.precision << endl;
  nv.push_back(n);

  int iter = 0;
  while (1) {

    double max = -1;
    int index = 0;
    for (int i = 0; i < nv.size(); i++) {
      if (nv[i].precision > max) {
        max = nv[i].precision;
        index = i;
      }
    }

    if (iter++ > 1000) {
      _is_used = nv[index].used_features;
      break;
    }

    
    Node n;
    n = nv[index];

    int count = 0;
    for (int i = 0; i < _num_features; i++)
      count += n.used_features[i];

    //    int num = rand() % (count * 0.01);
    int num = (int)(count * 0.01);
    ///    int num = 100;
    for (int i = 0; i < num; i++) {
      int j = rand() % _num_features;
      if (n.used_features[j]) count--;
      n.used_features[j] = false;
    }
    /*
    for (int i = 0; i < 10; i++) {
      int j = rand() % _num_features;
      n.used_features[j] = true;
    }
    */
    
    /*
    for (int i = 0; i < _num_features; i++) {
      if (n.used_features[i] == true) {
        n.used_features[i] = false;
        break;
      }
    }
    */

    _is_used = n.used_features;
    n.precision = fast_leave_one_out();

    nv.push_back(n);

    cout << count << " " << n.precision << endl;
  }

  

}

//------------------------------------------------------------------------
// Fast leave-one-out
//------------------------------------------------------------------------
double NB_Learner::fast_leave_one_out(vector<int>* pred_label_vec,
                                      bool noreduction)
{
  // calculate the auxilirary all_false_logp for each class
  vector<double> aux_af;

  for (int i = 0; i < _num_classes; i++) {
    aux_af.push_back(0);
  }
  
  for (int i = 0; i < _num_classes; i++) {
    // calculate all_false_logp
    double all_false_logp = 0;
    for (int j = 0; j < _num_features; j++) {
      if (!_is_used[j]) continue;  // ignore this feature
      double c = _cv[i][j];
      double p = c / _cc[i];
      if (p == 0) p = DELTA;
      if (p == 1) p = 1 - DELTA;
      _pv[i][j] = p;
      all_false_logp += log(1-p);
    }
    _af[i] = all_false_logp;

    if (_cc[i] <= 1) continue;

    // calculate all_false_logp for the reduced training data
    all_false_logp = 0;
    for (int j = 0; j < _num_features; j++) {
      if (!_is_used[j]) continue;  // ignore this feature
      double c = _cv[i][j];
      double p = c / (_cc[i] - 1);
      if (p == 0) p = DELTA;
      //if (p == 1) p = 1 - DELTA;
      all_false_logp += log(1-p);
    }
    aux_af[i] = all_false_logp;
  }

  // to calculate recall/precision for each class
  vector<int> class_correct(_num_classes);
  vector<int> class_labeled(_num_classes);
  vector<int> class_true(_num_classes);
  for (int i = 0; i < _num_classes; i++) {
    class_correct[i] = 0;
    class_labeled[i] = 0;
    class_true[i] = 0;
  }

  if (pred_label_vec) pred_label_vec->resize(_sv.size());

  // leave one out
  int correct = 0;
  int total = 0;
  for (int iter = 0; iter < _sv.size(); iter++) {
    const Sample& s = _sv[iter];
    assert(s.label >= 0);

    // evaluate the exmample
    double max = 0;
    int max_class = -1;
    for (int i = 0; i < _num_classes; i++) {
      double sum = 0;
      int cc = _cc[i];
      if (i == s.label && noreduction == false) {
        if (_cc[i] <= 1) continue;
        sum += log( ( (double)cc - 1 + 1) / (_sv.size() - 1 + _num_classes) );
        sum += aux_af[i];
        for (vector<int>::const_iterator j = s.positive_features.begin();
             j != s.positive_features.end(); j++) {
          if (*j >= _pv[i].size()) continue;   // unknown feature!
          if (!_is_used[*j]) continue;  // ignore this feature
          double c = _cv[i][*j];
          double p = c / (_cc[i] - 1);
          if (p == 0) p = DELTA;
          sum -= log(1 - p);
          double p2 = (c - 1) / (_cc[i] - 1);
          if (p2 == 0) p2 = DELTA;
          if (p2 == 1) p2 = 1 - DELTA;
          sum += log(p2);
        }
      } else {
        if (_cc[i] <= 0) continue;
        sum += log( ( (double)cc + 1) / (_sv.size() - 1 + _num_classes) );
        sum += _af[i];
        for (vector<int>::const_iterator j = s.positive_features.begin();
             j != s.positive_features.end(); j++) {
          if (*j >= _pv[i].size()) continue;   // unknown feature!
          if (!_is_used[*j]) continue;  // ignore this feature
          double p = _pv[i][*j];
          assert(p > 0 && p < 1);
          sum -= log(1 - p);
          sum += log(p);
        }
      }
      
      if (max_class < 0 || sum > max) {
        max = sum;
        max_class = i;
      }
    }

    int prediction = max_class;
    if (prediction == s.label) {
      correct++;
      class_correct[prediction]++;
    }
    total++;
    class_labeled[prediction]++;
    class_true[s.label]++;

    if (pred_label_vec) (*pred_label_vec)[iter] = prediction;

    //    cout << prediction << " " << s.label << endl;

    //    cerr << ".";

  }

  double precision = (double)correct / total;
  //  cout << "FAST precision: " << correct << "/"
  //       << total << " = " << precision << endl;

  /**
  // display precision/recall/f-score for each class
  for (int i = 0; i < _num_classes; i++) {
    cout << i << "\t";
    double precision = (double)class_correct[i]/class_labeled[i];
    double recall    = (double)class_correct[i]/class_true[i];
    cout << precision << " / " << recall << " / ";
    cout << 2 * precision * recall / (precision + recall) << endl;
  }
  **/
  
  return  precision;
}


//------------------------------------------------------------------------
// Leave-one-out
//------------------------------------------------------------------------
double NB_Learner::leave_one_out()
{
  int correct = 0;
  int total = 0;
  for (int iter = 0; iter < _sv.size(); iter++) {
    const Sample& s = _sv[iter];
    assert(s.label >= 0);

    // remove one example
    _cc[s.label]--;
    for (vector<int>::const_iterator j = s.positive_features.begin();
	 j != s.positive_features.end(); j++) {
      _cv[s.label][*j]--;
    }

    calc_probability();

    // evaluate the exmample
    NB_Sample nbs;
    nbs.positive_features = s.positive_features;
    int prediction = classify(nbs);
    if (prediction == s.label) correct++;
    total++;

    //    cerr << ".";

    // give back the example
    _cc[s.label]++;
    for (vector<int>::const_iterator j = s.positive_features.begin();
	 j != s.positive_features.end(); j++) {
      _cv[s.label][*j]++;
    }

  }

  double precision = (double)correct / total;
  cout << "LOO  precision: " << correct << "/"
       << total << " = " << precision << endl;

  return  precision;

}


//------------------------------------------------------------------------
// Training error
//------------------------------------------------------------------------
double NB_Learner::training_error()
{
  int correct = 0;
  int total = 0;
  calc_probability();
  for (int iter = 0; iter < _sv.size(); iter++) {
    const Sample& s = _sv[iter];
    assert(s.label >= 0);


    // evaluate the exmample
    NB_Sample nbs;
    nbs.positive_features = s.positive_features;
    int prediction = classify(nbs);
    if (prediction == s.label) correct++;
    total++;
  }

  double precision = (double)correct / total;
  cout << "Training data  precision: " << correct << "/"
       << total << " = " << precision << endl;

  return  precision;

}


//------------------------------------------------------------------------
// Cross validation test
//------------------------------------------------------------------------
void NB_Learner::cross_validation(int nfold)
{
  const int num_samples = _sv.size();

  vector<int> shuffled(num_samples);
  for (int i = 0; i < shuffled.size(); i++) {
    shuffled[i] = i;
  }
  for (int i = 0; i < shuffled.size(); i++) {
    int rest = shuffled.size() - i;
    int j = i + rand() % rest;

    // swap i, j
    int tmp = shuffled[j];
    shuffled[j] = shuffled[i];
    shuffled[i] = tmp;
  }

  int correct = 0;
  int total = 0;
  for (int iter = 0; iter < nfold; ++iter) {
    int test_begin = (num_samples / nfold) * iter;
    int test_end   = (num_samples / nfold) * (iter+1);

    // training
    vector<int> trains;
    for (int i = 0; i < num_samples; ++i) {
      if (i >= test_begin && i < test_end) continue; // test
      trains.push_back(i);
    }
    update(&trains);
    //    cerr << ".";

    // test
    for (int i = 0; i < num_samples; ++i) {
      if (i < test_begin || i >= test_end) continue; // train
      NB_Sample nbs;
      nbs.positive_features = _sv[i].positive_features;
      int prediction = classify(nbs);
      if (prediction == _sv[i].label) correct++;
      total++;
    }
  }
  cout << "precision: " << correct << "/"
       << total << " = " << (double)correct / total << endl;
    

}


//------------------------------------------------------------------------
// XiSquare independence test
//------------------------------------------------------------------------
void NB_Learner::independence_test()
{
  for (int i = 0; i < _num_features; i++) {
    for (int j = i+1; j < _num_features; j++) {
      vector<int> class_count(_num_classes);
      bool sufficient_num = true;
      int c_i = 0;
      for (int k = 0; k < _num_classes; k++) {
        int count = _cv[k][i];
        /*
        if (count < 3) {
          sufficient_num = false;
          break;
        }
        */
        c_i += count;
        class_count[k] += count;
      }
      if (!sufficient_num) continue;

      int c_j = 0;
      for (int k = 0; k < _num_classes; k++) {
        int count = _cv[k][j];
        /*
        if (count < 3) {
          sufficient_num = false;
          break;
        }
        */
        c_j += count;
        class_count[k] += count;
      }
      if (!sufficient_num) continue;

      double sum = 0;
      for (int k = 0; k < _num_classes; k++) {
        sum += _cv[k][i]*_cv[k][i] / class_count[k];
      }
      int total = c_i + c_j;
      sum -= c_i*c_i / total*total;
      sum *= total*total;
      sum /= (c_i*c_j);

      for (int k = 0; k < _num_classes; k++) {
        cout << _cv[k][i] << "\t";
      }
      cout << endl;
      for (int k = 0; k < _num_classes; k++) {
        cout << _cv[k][j] << "\t";
      }
      cout << endl;
      cout << "xi_square = " << sum << endl << endl;
    }
  }



}


//------------------------------------------------------------------------
// read samples from a file
//------------------------------------------------------------------------
/*******
void NB_Learner::read_from_file(const string& filename)
{
  _sv.clear();

  ifstream ifile(filename.c_str());
  if (!ifile) {
    throw runtime_error("cannot open" + filename);
  }

  string line;
  while (getline(ifile, line)) {
    vector<string> v = split(line);

    Sample s;

    vector<string>::const_iterator i = v.begin();
    int label = atoi(i->c_str());
    assert(label >= 0 && label < 1000);
    s.label = label;
    i++;
    for ( ;i != v.end(); i++) {
      int f = atoi(i->c_str());
      assert(f >= 0);
      s.positive_features.push_back(f);
      //      cout << f << " ";
    }

    _sv.push_back(s);
  }

  ifile.close();

  //  learn(_sv);
}
******/

//------------------------------------------------------------------------
// initialize the model
//------------------------------------------------------------------------
void NB_Learner::init()
{
  assert(_num_classes > 0 && _num_features > 0);

  DELTA = DEFAULT_DELTA;
  
  // initialize the model
  _pv.clear();
  _cv.clear();
  _cp.clear();
  _af.clear();
  _cc.clear();
  for (int i = 0; i < _num_classes; i++) {
    _cp.push_back(0.0);
    _af.push_back(0.0);
    _cc.push_back(0);

    vector<double> p;
    _pv.push_back(p);

    vector<int> c;
    _cv.push_back(c);
    
    for (int j = 0; j < _num_features; j++) {
      _pv[i].push_back(0.0);
      _cv[i].push_back(0);
    }
  }

}

//--------------------------------------------------------
// calculate the probability distribution of the model
//--------------------------------------------------------
void NB_Learner::calc_probability()
{
  int num_train_samples = 0;
  for (int i = 0; i < _num_classes; i++)
    num_train_samples += _cc[i];

  // calculate the probabilities
  for (int i = 0; i < _num_classes; i++) {
    if (_cc[i] == 0) continue;
    //    _cp[i] = (double)_cc[i] / num_train_samples;
    _cp[i] = ( (double)_cc[i] + 1) / (num_train_samples + _num_classes);
    double all_false_logp = 0;
    for (int j = 0; j < _num_features; j++) {
      if (!_is_used[j]) continue;  // ignore this feature
      //      double p = (double)(_cv[i][j] +1)/ (_cc[i] + 2);
      double c = _cv[i][j];
      double p = c / _cc[i];
      if (p == 0) p = DELTA;
      if (p == 1) p = 1 - DELTA;
      _pv[i][j] = p;
      all_false_logp += log(1-p);
    }
    _af[i] = all_false_logp;
  }
}


//------------------------------------------------------------------------
// update the model parameters
//    -- `trains' : the subset of _sv used for training
//------------------------------------------------------------------------
void NB_Learner::update(vector<int>* trains)
{
  init();

  vector<int> tmp_trains;
  if (!trains) {
    // use all the samples
    trains = &tmp_trains;
    for (int i = 0; i < _sv.size(); i++)
      trains->push_back(i);
  }

  // count
  for (vector<int>::const_iterator i = trains->begin(); i != trains->end();
       i++) {
    const Sample& s = _sv[*i];
    if (s.label == -1) continue;
    _cc[s.label]++;
    for (vector<int>::const_iterator j = s.positive_features.begin();
	 j != s.positive_features.end(); j++) {
      _cv[s.label][*j]++;
    }
  }

  calc_probability();
}


//------------------------------------------------------------------------
// calculate the model parameters
//------------------------------------------------------------------------
void NB_Learner::learn(const vector<NB_Sample>& tv)
{
  assert(tv.size() != 0);

  // check the number of classes and features
  int max_label = 0;
  int max_feature = 0;
  for (vector<NB_Sample>::const_iterator i = tv.begin(); i != tv.end(); i++) {
    if (i->label > max_label) max_label = i->label;
    for (vector<int>::const_iterator j = i->positive_features.begin();
	 j != i->positive_features.end(); j++) {
      if (*j > max_feature) max_feature = *j;
    }
  }
  _num_classes  = max_label   + 1;
  _num_features = max_feature + 1;

  // default: use all the features
  for (int j = 0; j < _num_features; j++) {
    _is_used.push_back(true);
  }

  // duplicate the given samples
  _sv.clear();
  for (vector<NB_Sample>::const_iterator i = tv.begin(); i != tv.end(); i++) {
    Sample s;
    s.label = i->label;
    s.positive_features = i->positive_features;
    //    sort(s.positive_features.begin(), s.positive_features.end());
    //    s.positive_features.unique();
    // sort & uniq ???
    _sv.push_back(s);
  }

  update();

  //  feature_selection();
  //  cross_validation(10);
  //  independence_test();

  //  cout << fast_leave_one_out() << endl;
}


//------------------------------------------------------------------------
// classify a NB_Sample
//------------------------------------------------------------------------
int NB_Learner::classify(const NB_Sample& nbs, double* logl_p,
                         vector<double>* memb_p)
{
  // add a sample at the end of the container
  Sample s;
  s.label = nbs.label;
  s.positive_features = nbs.positive_features;
  // sort & uniq ???
  _sv.push_back(s);

  int result = classify(_sv.size()-1, logl_p, memb_p);

  _sv.pop_back();
  return result;
}
  
//------------------------------------------------------------------------
// classify a sample in the container
//------------------------------------------------------------------------
int NB_Learner::classify(int index, double* logl_p,
                         vector<double>* memb_p)
{
  assert(memb_p == NULL || (*memb_p).size() == _num_classes);

  const Sample& s = _sv[index];
         
  double max = 0;
  int max_class = -1;
  double p_sum = 0;
  double likelihood = 0;
  for (int i = 0; i < _num_classes; i++) {
    if (_cc[i] == 0) {
      if (memb_p) (*memb_p)[i] = 0;
      continue;
    }
    double sum = 0;
    sum += log(_cp[i]);
    sum += _af[i];
    for (vector<int>::const_iterator j = s.positive_features.begin();
	 j != s.positive_features.end(); j++) {

      if (*j >= _pv[i].size()) continue;   // unknown feature!
      if (!_is_used[*j]) continue;  // ignore this feature
      
      double p = _pv[i][*j];
      assert(p > 0 && p < 1);
      sum -= log(1 - p);
      sum += log(p);
    }
    likelihood += exp(sum);
    if (memb_p != NULL) {
      double p = exp(sum);
      p_sum += p;
      (*memb_p)[i] = p;
    }
    if (max_class < 0 || sum > max) {
      max = sum;
      max_class = i;
    }
  }

  if (logl_p) *logl_p = log(likelihood);

  // normalize the class membership probabilities
  if (memb_p != NULL) {
    for (int i = 0; i < _num_classes; i++) {
      (*memb_p)[i] /= p_sum;
    }
  }

  //  cout << "logl = " << logl << endl;

  return max_class;
}


