#ifndef _NB_LEARNER_H_
#define _NB_LEARNER_H_

#include <vector>
#include <string>
//#include "mytools.h"

struct NB_Sample {
  int label;                          // -1: unlabeled
  std::vector<int> positive_features;
  NB_Sample() {};
  //  NB_Sample(const ML_Sample &s) {
  //    label = s.label;
  //    copy(s.af.begin(), s.af.end(), std::back_inserter(positive_features));
  //  }
};


class NB_Learner {

  struct Sample {
    int label;                          // -1: unlabeled
    std::vector<int> positive_features;
    std::vector<double> plabel;              // probabilistic label
  };
  
 public:
  NB_Learner() : _num_classes(0), _num_features(0) {}

  int num_classes() const { return _num_classes; }
  int num_features() const { return _num_features; }
  int num_samples() const { return _sv.size(); }

  void read_from_file(const std::string& filename);
  void learn(const std::vector<NB_Sample>& tv);
  int classify(const NB_Sample& s, double* logl_p = NULL,
               std::vector<double>* memb_p = NULL);
  void independence_test();
  void cross_validation(int nfold);
  double leave_one_out();
  double training_error();
  
  double fast_leave_one_out(std::vector<int>* pred_label_vec = NULL,
                            bool noreduction = false);
  void feature_selection();
  void optimize_delta();
  void set_delta(double delta) { DELTA = delta; };
  void feature_count(int count);
  
 private:
  double DELTA;  // the paramter to avoid zero-counting

  int _num_classes;
  int _num_features;
  std::vector<double> _cp;             // class probability
  std::vector<double> _af;             // all_false_logp
  std::vector<int>    _cc;             // count N(c_i)
  std::vector< std::vector<double> > _pv;
  std::vector< std::vector<int> >    _cv;   // count N(f_j, c_i)
  std::vector<Sample> _sv;
  std::vector<bool> _is_used;           // whether the feature is used or not

  void init();
  void calc_probability();
  void update(std::vector<int>* trains = NULL);
  int classify(int index, double* logl_p = NULL,
               std::vector<double>* memb_p = NULL);

};


#endif
